from subprocess import run

from logikal_utils.download import download
from logikal_utils.path import move, tmp_path, unzip
from selenium.webdriver.edge.options import Options
from selenium.webdriver.edge.service import Service
from selenium.webdriver.edge.webdriver import WebDriver

from logikal_browser import BrowserVersion
from logikal_browser.chromium import ChromiumBrowser


class EdgeVersion(BrowserVersion):
    name = 'edge'
    binary_name = 'microsoft-edge'
    driver_binary_name = 'msedgedriver'

    def install(self) -> None:
        if not self.path.exists():
            self.print(f'Installing Microsoft Edge {self.version}')
            root = 'https://packages.microsoft.com/repos/edge/pool/main/m/microsoft-edge-stable'
            url = f'{root}/microsoft-edge-stable_{self.version}-1_amd64.deb'
            tmp = tmp_path('logikal_browser', suffix=self.name)
            package = download(url, tmp / 'edge.deb')
            run(['dpkg', '-x', str(package), str(tmp / 'edge')], check=True)  # nosec
            move(tmp / 'edge/opt/microsoft/msedge', self.path.parent)

        if not self.driver_path.exists():
            self.print(f'Installing Microsoft Edge WebDriver {self.driver_version}')
            url = f'https://msedgedriver.microsoft.com/{self.version}/edgedriver_linux64.zip'
            tmp = tmp_path('logikal_browser', suffix=self.driver_name)
            unzip(download(url, tmp / 'edgedriver.zip'))
            move(tmp / 'edgedriver', self.driver_path.parent)


class EdgeBrowser(ChromiumBrowser, WebDriver):
    """
    Microsoft Edge WebDriver.
    """
    version_class = EdgeVersion
    options_class = Options
    service_class = Service

    # See https://github.com/SeleniumHQ/selenium/issues/14660
    height_offset = 78
    width_offset = 8



from logikal_utils.project import PYPROJECT

#: The available browser versions as specified in the ``tool.browser.versions`` section in
#: ``pyproject.toml``.
#:
#: .. tip:: The currently available versions can be found on the following URLs:
#:
#:  * `Chrome <https://googlechromelabs.github.io/chrome-for-testing/#stable>`_
#:  * `Edge <https://learn.microsoft.com/en-us/deployedge/microsoft-edge-relnote-stable-channel>`_
#:  * `Edge WebDriver <https://developer.microsoft.com/en-us/microsoft-edge/tools/webdriver>`_
BROWSER_VERSIONS = PYPROJECT.get('tool', {}).get('browser', {}).get('versions', {})



from collections.abc import Iterable
from dataclasses import dataclass


@dataclass
class Settings:
    """
    Browser settings specification.

    Args:
        name: Name of the settings.
        width: Browser window width.
        height: Browser window height.
        full_page_height: Whether to use the full page height for screenshots.
        mobile: Whether it is a mobile browser.

    """
    name: str
    width: int
    height: int
    full_page_height: bool = True
    mobile: bool = False


@dataclass
class Scenario:
    """
    Browser scenario specification.

    Args:
        settings: Settings to use.
        browsers: Browsers to use. Defaults to using all configured browser versions.

    """
    settings: Settings | Iterable[Settings]
    browsers: Iterable[str] | None = None


@dataclass
class StandardScenario(Scenario):
    """
    Standard browser scenario specification.
    """
    settings: Settings


desktop_4k = StandardScenario(Settings('desktop_4k', width=2560, height=1440))  #:
desktop = StandardScenario(Settings(
    'desktop',
    width=1920 - 120,  # offset for e.g. top bar, bottom bar, dash bar, task bar
    height=1080 - 180,  # offset for e.g. dash bar
))  #:
laptop_l = StandardScenario(Settings('laptop_l', width=1440, height=900))  #:
laptop = StandardScenario(Settings('laptop', width=1024, height=768))  #:
tablet = StandardScenario(Settings('tablet', width=768, height=1024, mobile=True))  #:
mobile_l = StandardScenario(Settings('mobile_l', width=425, height=680, mobile=True))  #:
mobile_m = StandardScenario(Settings('mobile_m', width=375, height=600, mobile=True))  #:
mobile_s = StandardScenario(Settings('mobile_s', width=320, height=512, mobile=True))  #:



from logikal_utils.download import download
from logikal_utils.path import move, tmp_path, unzip
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.webdriver import WebDriver

from logikal_browser import BrowserVersion
from logikal_browser.chromium import ChromiumBrowser


class ChromeVersion(BrowserVersion):
    name = 'chrome'

    def install(self) -> None:
        root = f'https://storage.googleapis.com/chrome-for-testing-public/{self.version}/linux64'

        if not self.path.exists():
            self.print(f'Installing Google Chrome {self.version}')
            tmp = tmp_path('logikal_browser', suffix=self.name)
            unzip(download(f'{root}/chrome-linux64.zip', tmp / 'chrome.zip'))
            move(tmp / 'chrome/chrome-linux64', self.path.parent)

        if not self.driver_path.exists():
            self.print(f'Installing Google Chrome WebDriver {self.driver_version}')
            tmp = tmp_path('logikal_browser', suffix=self.driver_name)
            unzip(download(f'{root}/chromedriver-linux64.zip', tmp / 'chromedriver.zip'))
            move(tmp / 'chromedriver/chromedriver-linux64', self.driver_path.parent)


class ChromeBrowser(ChromiumBrowser, WebDriver):
    """
    Google Chrome WebDriver.
    """
    version_class = ChromeVersion
    options_class = Options
    service_class = Service

    # See https://github.com/SeleniumHQ/selenium/issues/14660
    height_offset = 87



# pylint: disable=import-outside-toplevel
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from importlib import import_module
from pathlib import Path
from sys import stderr
from time import sleep
from typing import Any

from logikal_utils.path import tmp_path
from logikal_utils.testing import hide_traceback
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait
from termcolor import colored
from xdg_base_dirs import xdg_cache_home

from logikal_browser.config import BROWSER_VERSIONS
from logikal_browser.scenarios import Settings
from logikal_browser.utils import assert_image_equal

logger = logging.getLogger(__name__)


class BrowserVersion(ABC):
    """
    Base class for browser versions.

    Args:
        version: The browser version to use. Defaults to the appropriate version specified in
            :data:`~logikal_browser.config.BROWSER_VERSIONS`.
        install: Whether to install the web browser and the web driver.
        install_path: The path to use for the installation.
            Defaults to ``$XDG_CACHE_HOME/logikal_browser``.

    """
    def __init__(
        self,
        version: str | None = None,
        install: bool = True,
        install_path: Path | None = None,
    ):
        install_path = install_path or (xdg_cache_home() / 'logikal_browser')

        self.version = version or BROWSER_VERSIONS.get(self.name)
        if not self.version:
            raise ValueError(f'The browser version must be specified for "{self.name}"')
        self.path = install_path / self.name / self.version / self.binary_name
        self.driver_path = install_path / self.driver_name / self.version / self.driver_binary_name
        if install:
            self.install()

    def __str__(self) -> str:
        return f'{self.name} ({self.version})'

    def __repr__(self) -> str:
        return (
            f'<{str(self.__class__.__name__)} ({self.version}) at "{self.path}" '
            f'with {self.driver_name} ({self.driver_version}) at "{self.driver_path}">'
        )

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def install(self) -> None:
        ...

    @property
    def binary_name(self) -> str:
        return self.name

    @property
    def driver_name(self) -> str:
        return f'{self.name}driver'

    @property
    def driver_binary_name(self) -> str:
        return self.driver_name

    @property
    def driver_version(self) -> str:
        return self.version

    @staticmethod
    def print(message: str) -> None:
        print(colored(message, 'yellow', attrs=['bold']), file=stderr)


class Browser(ABC, WebDriver):
    """
    Base class for browser-specific web drivers.

    Args:
        settings: The browser settings to use.
        version: The browser version to use.
        headless: Whether to run in headless mode.
        screenshot_path: The path where screenshots are stored.
        screenshot_tmp_path: The temporary path to use for screenshots.
        download_path: The path to use for downloads.

    """
    height_offset = 0  # correction for https://github.com/SeleniumHQ/selenium/issues/14660
    width_offset = 0  # correction for https://github.com/SeleniumHQ/selenium/issues/14660

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        settings: Settings,
        version: BrowserVersion | None = None,
        headless: bool = True,
        screenshot_path: Path = Path('screenshot'),
        screenshot_tmp_path: Path | None = None,
        download_path: Path | None = None,
        **kwargs: Any,
    ):
        self.version = version or self.version_class()
        logger.debug(f'Browser version: {repr(self.version)}')
        self.settings = settings
        logger.debug(f'Browser settings: {self.settings}')
        self.headless = headless
        if self.headless:
            logger.debug('Running in headless mode')
        self.screenshot_path = screenshot_path
        logger.debug(f'Using screenshot path "{self.screenshot_path}/"')
        self.screenshot_tmp_path = (
            screenshot_tmp_path or tmp_path('logikal_browser', suffix='image')
        )
        logger.debug(f'Using screenshot temporary path "{self.screenshot_tmp_path}/"')
        self.download_path = download_path or tmp_path('logikal_browser', suffix='downloads')
        logger.debug(f'Using download path "{self.download_path}/"')

        super().__init__(**{**kwargs, **self.init_args()})

    @abstractmethod
    def init_args(self) -> dict[str, Any]:
        ...

    @property
    @abstractmethod
    def version_class(self) -> type[BrowserVersion]:
        ...

    def _set_settings_window_size(self, height: int | None = None) -> None:
        width = self.settings.width + self.width_offset
        height = (self.settings.height if height is None else height) + self.height_offset
        logger.debug(f'Setting window size to width "{width}" height "{height}"')
        self.set_window_size(width, height)

    @contextmanager
    def auto_height(self, wait_milliseconds: int | None) -> Iterator[None]:
        if not self.settings.full_page_height:
            yield
            return
        logger.debug('Using full page height')
        self._set_settings_window_size()
        if wait_milliseconds:  # we use a small delay to mitigate height flakiness
            logger.debug(f'Waiting {wait_milliseconds} ms')
            sleep(wait_milliseconds / 1000)
        elements = [
            'document.body.clientHeight',
            'document.body.scrollHeight',
            'document.body.offsetHeight',
            'document.documentElement.clientHeight',
            'document.documentElement.scrollHeight',
            'document.documentElement.offsetHeight',
        ]
        script = f'return Math.max({','.join(elements)});'
        height = self.execute_script(script)
        logger.debug(f'Calculated page height: {height}')
        self._set_settings_window_size(height=height)
        try:
            yield
        finally:
            self._set_settings_window_size()

    @hide_traceback
    def check(self, name: str | None = None, wait_milliseconds: int | None = 100) -> None:
        """
        Create a screenshot and check it against an expected version.

        Args:
            name: The name of the check.
            wait_milliseconds: The milliseconds to wait before calculating the screenshot height
                for unlimited height checks.

        """
        name_parts = [self.screenshot_path.name, self.version.name, self.settings.name, name]
        full_name = '_'.join(part for part in name_parts if part is not None)
        expected = self.screenshot_path.with_name(full_name).with_suffix('.png')

        script = 'document.body.style.caretColor = "transparent";'  # hide the blinking caret
        self.execute_script(script)

        with self.auto_height(wait_milliseconds=wait_milliseconds):
            logger.debug('Taking screenshot')
            # Note: we are disabling debug remote logs because they contain the verbose image data
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.INFO)
            actual = self.get_screenshot_as_png()
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.DEBUG)

        assert_image_equal(
            actual=actual,
            expected=expected,
            image_tmp_path=self.screenshot_tmp_path,
        )

    def replace_text(self, element: Any, text: str) -> None:
        """
        Replace the text of an element.

        Args:
            element: The element to use.
            text: The new text value.

        """
        script = f'arguments[0].innerHTML = "{text}";'
        self.execute_script(script, element)

    def wait_for_element(
        self,
        by: str,
        value: str,
        timeout_seconds: int = 10,
        poll_frequency: float = 0.5,
    ) -> None:
        """
        Wait until a given element is present.

        Args:
            by: The selector type to use for locating the element.
            value: The selector value to use for locating the element.
            timeout_seconds: The maximal time to wait.
            poll_frequency: Sleep interval between checks.

        """
        logger.debug(f'Waiting for element "{value}" to be present')
        wait = WebDriverWait(driver=self, timeout=timeout_seconds, poll_frequency=poll_frequency)
        wait.until(expected_conditions.presence_of_element_located((by, value)))

    def wait_for_download(
        self,
        name: str,
        timeout_seconds: int = 60,
        poll_frequency: float = 0.5,
    ) -> None:
        """
        Wait until a file with a given name is downloaded.

        Args:
            name: The name of the file to wait for.
            timeout_seconds: The maximal time to wait.
            poll_frequency: Sleep interval between checks.

        """
        file_path = self.download_path / name
        logger.debug(f'Waiting for file "{file_path}" to download')
        wait = WebDriverWait(driver=self, timeout=timeout_seconds, poll_frequency=poll_frequency)
        wait.until(lambda *_: file_path.exists())

    def login(self, user: Any, force: bool = True) -> None:
        """
        .. note:: The ``django`` extra must be installed for this method to work.

        Log in a given user.

        Args:
            user: The user to log in.
            force: Whether to log the user in without going through the authentication steps.

        """
        try:
            from django.conf import settings
            from django.contrib.auth import login as django_auth_login
            from django.http import HttpRequest
        except ImportError as error:  # pragma: no cover
            raise RuntimeError('The `django` extra must be installed for login to work') from error
        if not force:
            raise NotImplementedError('Only the forced login is implemented currently')

        request = HttpRequest()
        request.session = import_module(settings.SESSION_ENGINE).SessionStore()
        django_auth_login(request, user)
        request.session.save()
        self.add_cookie({
            'name': settings.SESSION_COOKIE_NAME,
            'value': request.session.session_key,
        })



#!/usr/bin/env python3
from logikal_browser.chrome import ChromeBrowser
from logikal_browser.scenarios import desktop

browser = ChromeBrowser(settings=desktop.settings, headless=False)
browser.get('https://logikal.io')
input('Press <ENTER> to stop the program... ')



import logging
import os
import sys
from io import BytesIO
from pathlib import Path
from shutil import copy
from subprocess import run

from logikal_utils.testing import hide_traceback
from PIL import Image, ImageChops

logger = logging.getLogger(__name__)


@hide_traceback
def assert_image_equal(actual: bytes, expected: Path, image_tmp_path: Path) -> None:
    tmp_actual_path = image_tmp_path / 'actual.png'
    tmp_expected_path = image_tmp_path / 'expected.png'
    tmp_diff_path = image_tmp_path / 'diff.png'

    logger.info(f'Checking expected image in "{expected}"')
    if expected.is_file():
        with Image.open(BytesIO(actual)) as actual_image, Image.open(expected) as expected_image:
            if actual_image == expected_image:
                logger.info('The actual image matches the expected image')
                return

            tmp_actual_path.write_bytes(actual)  # saving the temporary actual image
            copy(expected, tmp_expected_path)  # saving the temporary expected image
            diff_image = ImageChops.invert(ImageChops.difference(
                expected_image.convert('RGB'),
                actual_image.convert('RGB'),
            ))
            diff_image.save(str(tmp_diff_path))  # saving the temporary diff image
            save_image_prompt(
                message='Actual image differs from the expected image',
                source=tmp_actual_path, destination=expected, difference=tmp_diff_path,
            )
    else:
        logger.info('Expected image not found')
        tmp_actual_path.write_bytes(actual)  # saving the temporary actual image
        save_image_prompt(
            message='Expected image file does not exist',
            source=tmp_actual_path, destination=expected,
        )


def save_image_prompt(
    message: str,
    source: Path,
    destination: Path,
    difference: Path | None = None,
) -> None:
    if not sys.stdin.isatty():
        error_lines = [
            f'{message} and this is not an interactive session (consider using --live)',
            f'  Actual: {source.as_uri()}',
            f'  Expected: {destination.as_uri()}',
        ]
        if difference:
            error_lines.append(f'  Difference: {difference.as_uri()}')
        raise AssertionError('\n'.join(error_lines))
    try:
        short_destination = destination.relative_to(os.getcwd())
    except ValueError:
        short_destination = destination

    colors = {'red': '\033[31m\033[1m', 'reset': '\033[0m'}
    print(f'\n{colors['red']}{message}!{colors['reset']}')
    print(short_destination)

    prompt = f'{colors['red']}>{colors['reset']} '
    if (opener := Path('/usr/bin/xdg-open')).exists():
        response = input(f'{prompt}Press "enter" to open or type "s" to skip or "c" to cancel: ')
        if response == 's':
            logger.info('Image opening skipped')
        elif not response:
            # This subprocess call is secure as it is not using untrusted input
            run([str(opener), str(source)], check=False)  # nosec
        else:
            raise AssertionError('Image opening canceled')
    else:
        print(f'{prompt}See {source.as_uri()}')

    response = input(f'{prompt}Type "accept" to accept this version or press "enter" to reject: ')
    if response == 'accept':
        destination.parent.mkdir(parents=True, exist_ok=True)
        copy(source, destination)
        logger.info(f'Image saved at "{short_destination}"')
    else:
        raise AssertionError(f'Image rejected ({source.as_uri()})')



import os
from abc import abstractmethod
from typing import Any

from logikal_utils.random import DEFAULT_RANDOM_SEED
from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.chromium.service import ChromiumService

from logikal_browser import Browser


class ChromiumBrowser(Browser):
    """
    Abstract base class for Chromium-based WebDriver classes.
    """
    @property
    @abstractmethod
    def options_class(self) -> type[ChromiumOptions]:
        ...

    @property
    @abstractmethod
    def service_class(self) -> type[ChromiumService]:
        ...

    # See https://www.selenium.dev/documentation/webdriver/browsers/chrome/#options
    # See https://github.com/GoogleChrome/chrome-launcher/blob/main/docs/chrome-flags-for-tools.md
    def init_args(self) -> dict[str, Any]:
        window_width = self.settings.width + self.width_offset
        window_height = self.settings.height + self.height_offset
        args = [
            '--new-window',
            '--window-position=0,0',
            f'--window-size={window_width},{window_height}',
            # Unwanted features
            '--enable-automation',
            '--no-experiments',
            '--disable-infobars',
            '--disable-breakpad',
            '--disable-default-apps',
            '--disable-extensions',
            '--disable-component-extensions-with-background-pages',
            '--disable-features=InterestFeedContentSuggestions,Translate',
            '--no-default-browser-check',
            '--no-first-run',
            '--ash-no-nudges',
            '--disable-search-engine-choice-screen',
            '--propagate-iph-for-testing',
            # Deterministic rendering (see https://issues.chromium.org/issues/40039960#comment29)
            '--disable-partial-raster',
            '--disable-skia-runtime-opts',
            '--force-color-profile=srgb',
            f'--js-flags=--random-seed={DEFAULT_RANDOM_SEED}',
        ]
        if os.getenv('DOCKER_RUN') == '1':  # pragma: no cover
            args += ['--no-sandbox']
        if self.headless:
            args += [
                '--headless=new',
                '--in-process-gpu',  # memory saving
                '--mute-audio',
                '--allow-pre-commit-input',  # see https://issuetracker.google.com/issues/172339334
                '--deterministic-mode',
            ]

        options = self.options_class()
        options.binary_location = str(self.version.path)
        preferences = {'download.default_directory': str(self.download_path)}
        options.add_experimental_option('prefs', preferences)
        if self.settings.mobile:
            args.append('--hide-scrollbars')
        for arg in args:
            options.add_argument(arg)

        service = self.service_class(executable_path=str(self.version.driver_path))

        return {'options': options, 'service': service}



from logikal_browser import Browser, BrowserVersion
from logikal_browser.chrome import ChromeBrowser
from logikal_browser.config import BROWSER_VERSIONS
from logikal_browser.edge import EdgeBrowser


class InstalledBrowser:  # pylint: disable=too-few-public-methods
    """
    Information related to an installed browser.

    Args:
        browser_name: The name of the browser to install.
        version: The browser version to install.
        install: Whether to install the browser if necessary.

    """
    BROWSERS: dict[str, type[Browser]] = {
        'chrome': ChromeBrowser,
        'edge': EdgeBrowser,
    }

    def __init__(self, browser_name: str, version: str, install: bool):
        try:
            #: The appropriate browser-specific :class:`~logikal_browser.Browser` sub-class.
            self.browser_class: type[Browser] = InstalledBrowser.BROWSERS[browser_name]
        except KeyError as error:
            raise RuntimeError(f'Browser "{browser_name}" is currently not supported') from error

        browser_version_class: type[BrowserVersion] = (
            self.browser_class.version_class  # type: ignore[assignment]
        )
        #: The installed browser-specific :class:`~logikal_browser.BrowserVersion` sub-class
        #: instance.
        self.browser_version: BrowserVersion = browser_version_class(
            version=version,
            install=install,
        )


def installed_browsers(
    versions: dict[str, str] | None = None,
    install: bool = True,
) -> dict[str, InstalledBrowser]:
    """
    Return a mapping of browser names to installed browsers.

    Args:
        versions: A mapping of browser names to versions to install.
            Defaults to :data:`~logikal_browser.config.BROWSER_VERSIONS`.
        install: Whether to install the browsers if necessary.

    """
    installed: dict[str, InstalledBrowser] = {}

    if not (versions := BROWSER_VERSIONS if versions is None else versions):
        raise RuntimeError('You must specify at least one browser version')

    for browser_name, version in sorted(versions.items()):
        installed[browser_name] = InstalledBrowser(
            browser_name=browser_name,
            version=version,
            install=install,
        )

    return installed
